/* Advanced Laser Chess sound routines */

#include "advlaser.h"
#include "advsound.h"
#include <devices/audio.h>
#include <stdio.h>

/* Externals */
extern UWORD clup;  /* ADVDOOR.C */
extern int turn;    /* ADVLASER.C */
extern void *allocate_mem(); /* ADVLASER.C */

/* Local variables */
struct IOAudio *ioaudio;         /* Main ioaudio structure */
struct Unit *unit[4];            /* Individual channel units */
WORD allockey[4];                /* Individual channel allocation keys */
BOOL in_use[4];                  /* Keeps track of which channels in use */
struct MsgPort *static_port, *dynamic_port;
UBYTE *load_raw_sample();
UBYTE *waveform[NUMSAMPLES];
ULONG wavelength[NUMSAMPLES];
int general_channel = -1;
int background_on = FALSE, background_channel = -1, background_period;
int victory_period, victory_volume = 0;
int title_period, title_volume = 0;




init_sound()
{
  register int i;

  if((ioaudio = allocate_mem((long)sizeof(struct IOAudio),
  MEMF_PUBLIC | MEMF_CLEAR)) == NULL)
    cleanup(NOMEM);
  if((static_port = CreatePort(NULL, NULL)) == NULL)
    cleanup(NOAUDIOREPLYPORT);
  if((dynamic_port = CreatePort(NULL, NULL)) == NULL)
    cleanup(NOAUDIOREPLYPORT);
  ioaudio->ioa_Request.io_Message.mn_ReplyPort = static_port;
  if((OpenDevice("audio.device", NULL, ioaudio, NULL)) != NULL)
    cleanup(NOAUDIODEVICE);
  clup |= CLUP_AUDIODEV;

  for(i=0; i<4; i++)
  {
    allockey[i] = 0;
    unit[i] = NULL;
    in_use[i] = FALSE;
  }
  general_channel = allocate_audio_channel(GENERAL_CHANNEL_PRIORITY);
  load_waveforms();
}



beep()
{
  static int b = 0;
  register int i, at_least_one = FALSE;

  for(i=0; i<NUMBEEPS; i++)  /* See if any "beep" waveforms exist */
    if(waveform[i] != NULL)
      at_least_one = TRUE;
  if(!at_least_one) return; /* Return; no beeps are present to be played */

  /* Make sure that existing waveforms are played, as opposed to letting
  missing "beep" waveforms be played (as silence) */
  do
  {
    if(waveform[b] == NULL) ++b;
    if(b >= NUMBEEPS) b = 0;
  } while(waveform[b] == NULL);

  end_sound(general_channel, !NOW);
  switch(b)
  {
    case 0:
      start_sound(general_channel, waveform[WAVE_BEEP0],
        wavelength[WAVE_BEEP0], BEEP0_PERIOD, BEEP0_VOLUME, BEEP0_CYCLES);
      break;

    case 1:
      start_sound(general_channel, waveform[WAVE_BEEP1],
        wavelength[WAVE_BEEP1], BEEP1_PERIOD, BEEP1_VOLUME, BEEP1_CYCLES);
      break;
  }
  if(++b >= NUMBEEPS) b = 0;
}



disk_open_sound()
{
  end_sound(general_channel, NOW);
  start_sound(general_channel, waveform[WAVE_DISKOPEN],
    wavelength[WAVE_DISKOPEN], DISKOPEN_PERIOD, DISKOPEN_VOLUME,
    DISKOPEN_CYCLES);
}



disk_close_sound()
{
  end_sound(general_channel, NOW);
  start_sound(general_channel, waveform[WAVE_DISKCLOSE],
    wavelength[WAVE_DISKCLOSE], DISKCLOSE_PERIOD, DISKCLOSE_VOLUME,
    DISKCLOSE_CYCLES);
}



laser_sound(player)
register int player;
{
  register int p;

  p = player == RED? 0 : 200;
  laser_sound_off();
  start_sound(general_channel, waveform[WAVE_LASER], wavelength[WAVE_LASER],
    LASER_PERIOD + p, LASER_VOLUME, LASER_CYCLES);
}



laser_sound_off()
{
  end_sound(general_channel, NOW);
}



hyper_in_sound()
{
  end_sound(general_channel, NOW);
  start_sound(general_channel, waveform[WAVE_HYPERIN],
    wavelength[WAVE_HYPERIN], HYPERIN_PERIOD + rnd(150), HYPERIN_VOLUME,
    HYPERIN_CYCLES);
}



hyper_out_sound()
{
  end_sound(general_channel, NOW);
  start_sound(general_channel, waveform[WAVE_HYPEROUT],
    wavelength[WAVE_HYPEROUT], HYPEROUT_PERIOD + rnd(100), HYPEROUT_VOLUME,
    HYPEROUT_CYCLES);
}



crunch_sound()
{
  end_sound(general_channel, NOW);
  start_sound(general_channel, waveform[WAVE_CRUNCH],
    wavelength[WAVE_CRUNCH], CRUNCH_PERIOD, CRUNCH_VOLUME,
    CRUNCH_CYCLES);
}



explode_sound()
{
  end_sound(general_channel, NOW);
  start_sound(general_channel, waveform[WAVE_EXPLODE],
    wavelength[WAVE_EXPLODE], EXPLODE_PERIOD, EXPLODE_VOLUME, EXPLODE_CYCLES);
}



background_sound(on_off, volume)
register int on_off, volume;   /* TRUE=on, FALSE=off, TOGGLE=switch on/off */
{
  if(turn == RED) background_period = BACKGROUND_PERIOD;
  else background_period = BACKGROUND_PERIOD + 8;

  if(on_off == TOGGLE)
    background_on = !background_on;

  if(on_off == TRUE || (on_off == TOGGLE && background_on == TRUE))
  {
    if(background_channel == -1)
      background_channel = allocate_audio_channel(BACKGROUND_PRIORITY);
    start_sound(background_channel, waveform[WAVE_BACKGROUND],
      wavelength[WAVE_BACKGROUND], background_period, volume,
      BACKGROUND_CYCLES);
    background_on = TRUE;
  }
  else if(on_off == FALSE || (on_off == TOGGLE && background_on == FALSE))
  {
    end_sound(background_channel, NOW);
    free_audio_channel(background_channel);
    background_channel = -1;
    background_on = FALSE;
  }
}



modify_background_sound()  /* Called from fade() (ADVGRAPH.C) */
{
  background_period += turn == RED? -1 : 1;
  if(background_period < BACKGROUND_PERIOD)
    background_period = BACKGROUND_PERIOD;
  else if(background_period > BACKGROUND_PERIOD + 8)
    background_period = BACKGROUND_PERIOD + 8;

  modify_sound(background_channel, background_period, BACKGROUND_VOLUME);
}



victory_chord(on_off)  /* TRUE = on, FALSE = off */
register int on_off;
{
  static int back;

  end_sound(general_channel, NOW);
  if(on_off == TRUE)
  {
    back = background_on;
    if(background_on) background_sound(FALSE, 0);
    /* Start victory chord at volume zero */
    victory_volume = 0;
    victory_period = VICTORY_PERIOD;
    start_sound(general_channel, waveform[WAVE_VICTORY],
      wavelength[WAVE_VICTORY], VICTORY_PERIOD, 0,
      VICTORY_CYCLES);
  }
  else  /* If turning off sound, reset victory sound parameters */
  {
    end_sound(general_channel, NOW);
    victory_period = VICTORY_PERIOD;
    victory_volume = 0;
    if(back) background_sound(TRUE, BACKGROUND_VOLUME);
  }
}



modify_victory_sound(up_down)  /* up_down = 1 for up, -1 for down. */
register int up_down;
{
  victory_period -= 5 * up_down;
  victory_volume += 2 * up_down;

  modify_sound(general_channel, victory_period, victory_volume);
}



title_sound(on_off)  /* TRUE=on, FALSE=off */
{
  static int back;

  if(on_off == TRUE)
  {
    back = background_on;
    if(background_on) background_sound(FALSE, 0);

    background_channel = allocate_audio_channel(BACKGROUND_PRIORITY);

    /* Start title chords at zero volume */
    title_volume = 0;
    title_period = TITLE_PERIOD;
    start_sound(general_channel, waveform[WAVE_TITLE],
      wavelength[WAVE_TITLE], TITLE_PERIOD, 0, TITLE_CYCLES);
    start_sound(background_channel, waveform[WAVE_TITLE],
      wavelength[WAVE_TITLE], TITLE_PERIOD, 0, TITLE_CYCLES);
  }
  else  /* If turning off sound, reset title sound parameters */
  {
    end_sound(general_channel, NOW);
    end_sound(background_channel, NOW);
    title_period = TITLE_PERIOD;
    title_volume = 0;
    if(back)
      background_sound(TRUE, 0);
  }
}



modify_title_sound(up_down)  /* up_down = 1 for up, -1 for down. */
register int up_down;
{
  title_volume += 1 * up_down;
  title_period -= 1 * up_down;
  if(title_volume < 0) title_volume = 0;

  modify_sound(general_channel, title_period,
    title_volume > TITLE_MAX_VOLUME? TITLE_MAX_VOLUME : title_volume);
  modify_sound(background_channel, title_period,
    title_volume > TITLE_MAX_VOLUME? TITLE_MAX_VOLUME : title_volume);
}



rnd_title_chord(first)  /* Changes the title chord */
register int first;
{
  static int period1[] = {306, 274, 306, 229, 306, 171, 388, 171, 306};
  static int period2[] = {306, 306, 256, 306, 204, 345, 153, 345, 153};
  static int count = 0, note_count = 0;

  if(first)
    note_count = 0;

  if(++count < 12000)
    return;
  count = 0;

  modify_sound(general_channel, period1[note_count],
    title_volume > TITLE_MAX_VOLUME? TITLE_MAX_VOLUME : title_volume);

  modify_sound(background_channel, period2[note_count],
    title_volume > TITLE_MAX_VOLUME? TITLE_MAX_VOLUME : title_volume);

  if(++note_count >= 9)
    note_count = 8;
}



load_waveforms()
{
  register int i, incomplete;
  static char *notall[] =
  {
    "Not all of the sound effects",
    "could be loaded.  If you wish",
    "to hear all possible sounds,",
    "quit this program, free up some",
    "memory (by closing existing",
    "programs or deleting files from",
    "RAM), then run this program again."
  };
  static char *none[] =
  {
    "NONE of the sound effects",
    "could be loaded.  Make sure",
    "that the \"samples\" directory",
    "exists on the same disk as",
    "Advanced Laser Chess and rerun",
    "the program if you wish to",
    "hear all of the sound effects."
  };

  /* The order in which the waveforms are loaded denote their priority.
  That is, the first waveform loaded (considered the most important)
  has the best chance of successfully "making it" into memory.  The
  succeeding waveforms have a lower chance of successfully loading
  because of the continuous memory decline. */

  waveform[WAVE_BACKGROUND] = load_raw_sample(BACKGROUND_FILENAME,
    &wavelength[WAVE_BACKGROUND]);

  waveform[WAVE_BEEP0] = load_raw_sample(BEEP0_FILENAME,
    &wavelength[WAVE_BEEP0]);

  waveform[WAVE_BEEP1] = load_raw_sample(BEEP1_FILENAME,
    &wavelength[WAVE_BEEP1]);

  waveform[WAVE_LASER] = load_raw_sample(LASER_FILENAME,
    &wavelength[WAVE_LASER]);

  waveform[WAVE_CRUNCH] = load_raw_sample(CRUNCH_FILENAME,
    &wavelength[WAVE_CRUNCH]);

  waveform[WAVE_EXPLODE] = load_raw_sample(EXPLODE_FILENAME,
    &wavelength[WAVE_EXPLODE]);

  waveform[WAVE_VICTORY] = load_raw_sample(VICTORY_FILENAME,
    &wavelength[WAVE_VICTORY]);

  waveform[WAVE_TITLE] = load_raw_sample(TITLE_FILENAME,
    &wavelength[WAVE_TITLE]);

  waveform[WAVE_HYPERIN] = load_raw_sample(HYPERIN_FILENAME,
    &wavelength[WAVE_HYPERIN]);

  waveform[WAVE_HYPEROUT] = load_raw_sample(HYPEROUT_FILENAME,
    &wavelength[WAVE_HYPEROUT]);

  waveform[WAVE_DISKOPEN] = load_raw_sample(DISKOPEN_FILENAME,
    &wavelength[WAVE_DISKOPEN]);

  waveform[WAVE_DISKCLOSE] = load_raw_sample(DISKCLOSE_FILENAME,
    &wavelength[WAVE_DISKCLOSE]);

  incomplete = TRUE;  /* Check to see if ANY waveforms got loaded */
  for(i=0; i<NUMSAMPLES; i++)
    if(waveform[i] != NULL)
      incomplete = FALSE;
  if(incomplete)  /* If none loaded, display message */
    req(none, "Ok", "Ok", 7, FALSE);
  else  /* Check to see if ALL of the waveforms have been loaded */
  {
    incomplete = FALSE;
    for(i=0; i<NUMSAMPLES; i++)
      if(waveform[i] == NULL)
        incomplete = TRUE;
    if(incomplete)  /* If not, display message */
      req(notall, "Ok", "Ok", 7, FALSE);
  }
}



struct IOAudio *allocate_ioaudio(channel)
register int channel;
{
  struct IOAudio *new;

  if(channel < 0 || channel > 3 || !in_use[channel]) return;

  new = allocate_mem((long)sizeof(struct IOAudio), MEMF_PUBLIC | MEMF_CLEAR);
  if(new != NULL)
  {
    new->ioa_Request.io_Unit = unit[channel];
    new->ioa_Request.io_Device = ioaudio->ioa_Request.io_Device;
    new->ioa_Request.io_Message.mn_ReplyPort = dynamic_port;
    new->ioa_AllocKey = allockey[channel];
  }
  return(new);
}



/* Returns -1 if no channel available, otherwise 0 - 3. */
allocate_audio_channel(priority)
register int priority;
{
  static UBYTE audio_alloc[] = {1, 2, 8, 4}; /* Audio channel alloc. table */
  int channel_alloc;

  ioaudio->ioa_Request.io_Command = ADCMD_ALLOCATE;
  ioaudio->ioa_Request.io_Flags = ADIOF_NOWAIT;
  ioaudio->ioa_Request.io_Unit = NULL;
  ioaudio->ioa_AllocKey = 0;
  ioaudio->ioa_Request.io_Message.mn_Node.ln_Pri = priority;
  ioaudio->ioa_Data = audio_alloc;
  ioaudio->ioa_Length = 4L;

  BeginIO(ioaudio);
  if(ioaudio->ioa_Request.io_Error != NULL)
    return(-1);

  switch((long)ioaudio->ioa_Request.io_Unit)
  {
    case 1L: channel_alloc = 0; break;
    case 2L: channel_alloc = 1; break;
    case 4L: channel_alloc = 2; break;
    case 8L: channel_alloc = 3; break;
    default: channel_alloc = -1; break;
  }
  if(channel_alloc != -1)
  {
    unit[channel_alloc] = ioaudio->ioa_Request.io_Unit;
    allockey[channel_alloc] = ioaudio->ioa_AllocKey;
    in_use[channel_alloc] = TRUE;
  }

  return(channel_alloc);
}



start_sound(channel, data, length, period, volume, cycles)
int channel, period, volume, cycles;
UBYTE *data;
ULONG length;
{
  struct IOAudio *ioa;

  if(channel < 0 || channel > 3 || data == NULL || !in_use[channel])
    return;
  if((ioa = allocate_ioaudio(channel)) == NULL)
    return;

  ioa->ioa_Request.io_Command = CMD_WRITE;
  ioa->ioa_Request.io_Flags = ADIOF_PERVOL;
  ioa->ioa_Data = data;
  ioa->ioa_Length = length;
  ioa->ioa_Period = (long)period;
  ioa->ioa_Volume = (long)volume;
  ioa->ioa_Cycles = (long)cycles;
  clean_port();
  BeginIO(ioa);
}



modify_sound(channel, period, volume)
register int channel, period, volume;
{
  struct IOAudio ioa;

  if(channel < 0 || channel > 3 || !in_use[channel])
    return;

  /* QuickIO always succeeds for ADCMD_PERVOL, so I'll take advantage
  of the advantage and USE it; an IOAudio structure will not have to
  be dynamically allocated for each ADCMD_PERVOL for this reason. */
  ioa.ioa_Request.io_Unit = unit[channel];
  ioa.ioa_Request.io_Device = ioaudio->ioa_Request.io_Device;
  ioa.ioa_Request.io_Message.mn_ReplyPort = static_port;
  ioa.ioa_AllocKey = allockey[channel];
  ioa.ioa_Request.io_Command = ADCMD_PERVOL;
  ioa.ioa_Request.io_Flags = IOF_QUICK;
  ioa.ioa_Period = (long)period;
  ioa.ioa_Volume = (long)volume;
  BeginIO(&ioa);
}



end_sound(channel, sync)
register int channel, sync;
{
  struct IOAudio *ioa;

  if(channel < 0 || channel > 3 || !in_use[channel]) return;
  if((ioa = allocate_ioaudio(channel)) == NULL) return;

  ioa->ioa_Request.io_Command = ADCMD_FINISH;
  ioa->ioa_Request.io_Flags = sync == NOW? 0L : ADIOF_SYNCCYCLE;
  clean_port();
  BeginIO(ioa);
}



free_audio_channel(channel)
register int channel;
{
  if(channel < 0 || channel > 3 || !in_use[channel]) return;

  in_use[channel] = FALSE;

  ioaudio->ioa_Request.io_Unit = unit[channel];
  ioaudio->ioa_AllocKey = allockey[channel];
  ioaudio->ioa_Request.io_Command = ADCMD_FREE;
  ioaudio->ioa_Request.io_Flags = 0L;
  BeginIO(ioaudio);
}



free_waveforms()
{
  register int i;

  for(i=0; i<NUMSAMPLES; i++)
    if(waveform[i] != NULL)
      FreeMem(waveform[i], wavelength[i]);
}



finish_channels() /* Performs an ADCMD_FINISH command for each channel... */
{                 /* currently playing. */
  register int i;

  for(i=0; i<4; i++)
    free_audio_channel(i);
}



clean_port()  /* Frees all dynamic IOAudio structures from dynamic_port */
{
  struct IOAudio *ioa;

  while((ioa = (struct IOAudio *)GetMsg(dynamic_port)) != NULL)
    FreeMem(ioa, (long)sizeof(struct IOAudio));
}



UBYTE *load_raw_sample(filename, length)
char *filename;
ULONG *length;
{
  FILE *file;
  UBYTE *buffer;
  ULONG size;
  register ULONG i;

  if((file = fopen(filename, "r")) == NULL)
    return(NULL);

  fseek(file, 0L, 2);  /* Seek to end of file */
  size = ftell(file) & ~1L;  /* Size of file (in bytes) */
  fseek(file, 0L, 0);  /* Seek to beginning of file */

  /* Allocate a buffer the same size as the file about to be read */
  if((buffer = allocate_mem(size, MEMF_CHIP)) == NULL)
  {
    fclose(file);
    return(NULL);
  }

  /* NOTE: Files over 32K bytes cannot be handled using the following
  fread() call as it stands; if bigger files need to be loaded, change
  the fread() to load in longwords. */
  *length = (ULONG)fread(buffer, 1, (int)size, file);

  fclose(file);

  return(buffer);
}



close_sound()
{
  register int i;

  background_sound(FALSE, 0);
  finish_channels();
  if(dynamic_port != NULL)
    clean_port();
  if(clup & CLUP_AUDIODEV)
    CloseDevice(ioaudio);
  if(ioaudio != NULL)
    FreeMem(ioaudio, (long)sizeof(struct IOAudio));
  if(static_port != NULL)
    DeletePort(static_port);
  if(dynamic_port != NULL)
    DeletePort(dynamic_port);
  free_waveforms();
}


